// file      : xsd/cxx/parser/name-processor.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/parser/name-processor.hxx>

#include <cxx/parser/elements.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <cult/containers/set.hxx>

namespace CXX
{
  namespace Parser
  {
    namespace
    {
      //
      //
      typedef Cult::Containers::Set<String> NameSet;

      String
      find_name (SemanticGraph::Nameable& n, NameSet& set)
      {
        String base_name (Context::escape (n.name ()));
        String name (base_name);

        for (UnsignedLong i (1); set.find (name) != set.end (); ++i)
        {
          std::wostringstream os;
          os << i;
          name = base_name + os.str ();
        }

        set.insert (name);
        n.context ().set ("name", name);

        return name;
      }


      /*
      //
      //
      struct Enumerator: Traversal::Enumerator
      {
        Enumerator (NameSet& set)
            : set_ (set)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          find_name (e, set_);
        }

      private:
        NameSet& set_;
      };


      //
      //
      struct Enumeration: Traversal::Enumeration
      {
        Enumeration (String const& name_ = "")
            : name (name_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (e.named ())
            name = e.name ();

          NameSet enumerator_set;
          enumerator_set.insert (name);

          Enumerator enumerator (enumerator_set);
          Traversal::Names names;

          names >> enumerator;

          Traversal::Enumeration::names (e, names);
        }

      private:
        String name;
      };
      */

      //
      //
      struct Member: Traversal::Member
      {
        Member (NameSet& set)
            : set_ (set)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (Context::skip (m))
            return;

          find_name (m, set_);
        }

      private:
        NameSet& set_;
      };


      //
      //
      struct Complex: Traversal::Complex
      {
        Complex (String const& name_ = "")
            : name (name_)
        {
        }

        virtual Void
        traverse (Type& c)
        {
          if (c.named ())
            name = c.name ();

          NameSet member_set;
          member_set.insert (name);

          Member member (member_set);
          Traversal::Names names;

          names >> member;

          Traversal::Complex::names (c, names);
        }

      private:
        String name;
      };


      //
      //
      struct GlobalType: Traversal::Type
      {
        GlobalType (NameSet& set)
            : set_ (set)
        {
        }

        virtual Void
        traverse (SemanticGraph::Type& t)
        {
          find_name (t, set_);
        }

      private:
        NameSet& set_;
      };

      //
      //
      /*
      struct GlobalElement: Traversal::Element
      {
        GlobalElement (NameSet& set)
            : set_ (set)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          String name (find_name (e, set_));
        }

      private:
        NameSet& set_;
      };
      */

      struct Namespace: Traversal::Namespace
      {
        virtual Void
        traverse (Type& ns)
        {
          NameSet type_set;
          NameSet element_set;

          GlobalType type (type_set);
          //GlobalElement element (element_set);
          Traversal::Names names;

          names >> type;
          //names >> element;

          Traversal::Namespace::names (ns, names);
          Traversal::Namespace::names (ns);
        }
      };


      // Go into implied/included/imported schemas while making sure
      // we don't recurse too far.
      //
      struct Uses: Traversal::Uses
      {
        virtual Void
        traverse (Type& u)
        {
          SemanticGraph::Schema& s (u.schema ());

          if (!s.context ().count ("seen"))
          {
            s.context ().set ("seen", true);
            Traversal::Uses::traverse (u);
            s.context ().remove ("seen");
          }
        }
      };
    }

    Void NameProcessor::
    process (SemanticGraph::Schema& tu,
             SemanticGraph::Path const& file)
    {
      Traversal::Schema schema;
      Uses uses;

      schema >> uses >> schema;

      Traversal::Names schema_names;
      Namespace ns;
      Traversal::Names ns_names;

      schema >> schema_names >> ns >> ns_names;

      Complex complex;
      //Enumeration enumeration;

      ns_names >> complex;
      //ns_names >> enumeration;

      schema.dispatch (tu);
    }
  }
}
