// file      : xsd/cxx/tree/parser-header.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/parser-header.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct Element : Traversal::Element,
                       GlobalElementBase,
                       protected virtual Context
      {
        Element (Context& c)
            : GlobalElementBase (c),
              Context (c),
              esymbol (c.esymbol.empty () ? c.esymbol : c.esymbol + L"\n")
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (!doc_root_p (e))
            return;

          String const& name (ename (e));
          String properties_type (L"::xsd::cxx::tree::properties< " +
                                  char_type + L" >");

          os << "// Read from a URI or a local file." << endl
             << "//" << endl
             << endl;

          os << esymbol
             << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << name << " (" << string_type << " const&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << properties_type << " const& = " << properties_type << " ());"
             << endl;

          os << esymbol
             << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << name << " (" << string_type << " const&," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << properties_type << " const& = " << properties_type << " ());"
             << endl;

          os << esymbol
             << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << name << " (" << string_type << " const&," << endl
             << "::xercesc::DOMErrorHandler&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << properties_type << " const& = " << properties_type << " ());"
             << endl
             << endl;


          os << "// Read from std::istream." << endl
             << "//" << endl
             << endl;

          os << esymbol
             << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << name << " (::std::istream&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << properties_type << " const& = " << properties_type << " ());"
             << endl;

          os << esymbol
             << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << name << " (::std::istream&," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << properties_type << " const& = " << properties_type << " ());"
             << endl;

          os << esymbol
             << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << name << " (::std::istream&," << endl
             << "::xercesc::DOMErrorHandler&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << properties_type << " const& = " << properties_type << " ());"
             << endl
             << endl;

          os << esymbol
             << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << name << " (::std::istream&," << endl
             << string_type << " const& id," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << properties_type << " const& = " << properties_type << " ());"
             << endl;

          os << esymbol
             << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << name << " (::std::istream&," << endl
             << string_type << " const& id," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << properties_type << " const& = " << properties_type << " ());"
             << endl;

          os << esymbol
             << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << name << " (::std::istream&," << endl
             << string_type << " const& id," << endl
             << "::xercesc::DOMErrorHandler&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << properties_type << " const& = " << properties_type << " ());"
             << endl
             << endl;

          /*
          os << esymbol
             << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << name << " (::std::istream&, " <<
            string_type << " const& system_id, " <<
            string_type << " const& public_id);"
             << endl;

          os << esymbol
             << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << name << " (::std::istream&, " <<
            string_type << " const& system_id, " <<
            string_type << " const& public_id, " <<
            "::xercesc::DOMErrorHandler&);"
             << endl
             << endl;
          */


          os << "// Read from InputSource." << endl
             << "//" << endl
             << endl;

          os << esymbol
             << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << name << " (::xercesc::DOMInputSource const&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << properties_type << " const& = " << properties_type << " ());"
             << endl;

          os << esymbol
             << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << name << " (::xercesc::DOMInputSource const&," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << properties_type << " const& = " << properties_type << " ());"
             << endl;

          os << esymbol
             << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << name << " (::xercesc::DOMInputSource const&," << endl
             << "::xercesc::DOMErrorHandler&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << properties_type << " const& = " << properties_type << " ());"
             << endl
             << endl;


          os << "// Read from DOM." << endl
             << "//" << endl
             << endl;

          os << esymbol
             << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << name << " (::xercesc::DOMDocument const&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << properties_type << " const& = " << properties_type << " ());"
             << endl
             << endl;
        }

      private:
        String
        type_name (Type& e)
        {
          SemanticGraph::Type& t (e.type ());

          if (t.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (t);

            return o.str ();
          }
          else
          {
            String const& name (ename (e));
            return name + L"::" + unclash (name, "type");
          }
        }

      private:
        String esymbol;
      };
    }


    Void
    generate_parser_header (Context& ctx, SemanticGraph::Schema& tu)
    {
      ctx.os << "#include <iosfwd>" << endl
             << endl;

      Traversal::Schema schema;
      Traversal::Sources sources;
      Traversal::Names names_ns, names;
      Namespace ns (ctx, false);
      Element element (ctx);

      schema >> sources >> schema;
      schema >> names_ns >> ns >> names >> element;

      schema.dispatch (tu);
    }
  }
}
