// file      : xsd/cxx/tree/tree-header.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/tree-header.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <backend-elements/regex.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct FundamentalNamespace : Namespace,

                                    Traversal::AnyType,
                                    Traversal::AnySimpleType,

                                    Traversal::Fundamental::Byte,
                                    Traversal::Fundamental::UnsignedByte,
                                    Traversal::Fundamental::Short,
                                    Traversal::Fundamental::UnsignedShort,
                                    Traversal::Fundamental::Int,
                                    Traversal::Fundamental::UnsignedInt,
                                    Traversal::Fundamental::Long,
                                    Traversal::Fundamental::UnsignedLong,
                                    Traversal::Fundamental::Integer,
                                    Traversal::Fundamental::NonPositiveInteger,
                                    Traversal::Fundamental::NonNegativeInteger,
                                    Traversal::Fundamental::PositiveInteger,
                                    Traversal::Fundamental::NegativeInteger,

                                    Traversal::Fundamental::Boolean,

                                    Traversal::Fundamental::Float,
                                    Traversal::Fundamental::Double,
                                    Traversal::Fundamental::Decimal,

                                    Traversal::Fundamental::String,
                                    Traversal::Fundamental::NormalizedString,
                                    Traversal::Fundamental::Token,
                                    Traversal::Fundamental::Name,
                                    Traversal::Fundamental::NameToken,
                                    Traversal::Fundamental::NameTokens,
                                    Traversal::Fundamental::NCName,
                                    Traversal::Fundamental::Language,

                                    Traversal::Fundamental::QName,

                                    Traversal::Fundamental::Id,
                                    Traversal::Fundamental::IdRef,
                                    Traversal::Fundamental::IdRefs,

                                    Traversal::Fundamental::AnyURI,

                                    Traversal::Fundamental::Base64Binary,
                                    Traversal::Fundamental::HexBinary,

                                    Traversal::Fundamental::Date,
                                    Traversal::Fundamental::DateTime,
                                    Traversal::Fundamental::Duration,
                                    Traversal::Fundamental::Day,
                                    Traversal::Fundamental::Month,
                                    Traversal::Fundamental::MonthDay,
                                    Traversal::Fundamental::Year,
                                    Traversal::Fundamental::YearMonth,
                                    Traversal::Fundamental::Time,

                                    Traversal::Fundamental::Entity,
                                    Traversal::Fundamental::Entities,
                                    protected virtual Context
      {
        FundamentalNamespace (Context& c)
            : Namespace (c, true), Context (c)
        {
          *this >> names_ >> *this;
        }

        // anyType and anySimpleType
        //
        virtual Void
        traverse (SemanticGraph::AnyType&)
        {
          os << "// anyType and anySimpleType." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::tree::type type;";
        }

        virtual Void
        traverse (SemanticGraph::AnySimpleType&)
        {
          os << "typedef ::xsd::cxx::tree::simple_type simple_type;"
             << endl;
        }

        // Integrals.
        //
        virtual Void
        traverse (SemanticGraph::Fundamental::Byte&)
        {
          os << "// 8-bit" << endl
             << "//" << endl
             << "typedef signed char byte;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::UnsignedByte&)
        {
          os << "typedef unsigned char unsigned_byte;"
             << endl;
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Short&)
        {
          os << "// 16-bit" << endl
             << "//" << endl
             << "typedef short short_;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::UnsignedShort&)
        {
          os << "typedef unsigned short unsigned_short;"
             << endl;
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Int&)
        {
          os << "// 32-bit" << endl
             << "//" << endl
             << "typedef int int_;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::UnsignedInt&)
        {
          os << "typedef unsigned int unsigned_int;"
             << endl;
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Long&)
        {
          os << "// 64-bit" << endl
             << "//" << endl
             << "typedef long long long_;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::UnsignedLong&)
        {
          os << "typedef unsigned long long unsigned_long;"
             << endl;
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Integer&)
        {
          os << "// Supposed to be arbitrary-length integral types." << endl
             << "//" << endl
             << "typedef long long integer;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::NonPositiveInteger&)
        {
          os << "typedef integer non_positive_integer;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::NonNegativeInteger&)
        {
          os << "typedef integer non_negative_integer;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::PositiveInteger&)
        {
          os << "typedef integer positive_integer;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::NegativeInteger&)
        {
          os << "typedef integer negative_integer;"
             << endl;
        }

        // Boolean.
        //
        virtual Void
        traverse (SemanticGraph::Fundamental::Boolean&)
        {
          os << "// Boolean." << endl
             << "//" << endl
             << "typedef bool boolean;"
             << endl;
        }

        // Floats.
        //
        virtual Void
        traverse (SemanticGraph::Fundamental::Float&)
        {
          os << "// Floating-point types." << endl
             << "//" << endl
             << "typedef float float_;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Double&)
        {
          os << "typedef double double_;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Decimal&)
        {
          os << "typedef long double decimal;"
             << endl;
        }


        // Strings.
        //
        virtual Void
        traverse (SemanticGraph::Fundamental::String&)
        {
          os << "// String types." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::tree::string< " <<
            char_type << " > string;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::NormalizedString&)
        {
          os << "typedef ::xsd::cxx::tree::normalized_string< " <<
            char_type << " > normalized_string;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Token&)
        {
          os << "typedef ::xsd::cxx::tree::token< " <<
            char_type << " > token;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::NameToken&)
        {
          os << "typedef ::xsd::cxx::tree::nmtoken< " <<
            char_type << " > nmtoken;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::NameTokens&)
        {
          os << "typedef ::xsd::cxx::tree::nmtokens< " <<
            char_type << " > nmtokens;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Name&)
        {
          os << "typedef ::xsd::cxx::tree::name< " <<
            char_type << " > name;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::NCName&)
        {
          os << "typedef ::xsd::cxx::tree::ncname< " <<
            char_type << " > ncname;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Language&)
        {
          os << "typedef ::xsd::cxx::tree::language< " <<
            char_type << " > language;"
             << endl;
        }


        // Qualified name.
        //
        virtual Void
        traverse (SemanticGraph::Fundamental::QName&)
        {
          os << "// Qualified name." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::tree::qname< " <<
            char_type << " > qname;"
             << endl;
        }


        // ID/IDREF.
        //
        virtual Void
        traverse (SemanticGraph::Fundamental::Id&)
        {
          os << "// ID/IDREF." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::tree::id< " <<
            char_type << " > id;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::IdRef&)
        {
          os << "typedef ::xsd::cxx::tree::idref< type, " <<
            char_type << " > idref;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::IdRefs&)
        {
          os << "typedef ::xsd::cxx::tree::idrefs< type, " <<
            char_type << " > idrefs;"
             << endl;
        }


        // URI.
        //
        virtual Void
        traverse (SemanticGraph::Fundamental::AnyURI&)
        {
          os << "// URI." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::tree::uri< " <<
            char_type << " > uri;"
             << endl;
        }


        // Binary.
        //
        virtual Void
        traverse (SemanticGraph::Fundamental::Base64Binary&)
        {
          os << "// Binary." << endl
             << "//" << endl
            << "typedef ::xsd::cxx::tree::buffer< " <<
            char_type << " > buffer;"
             << "typedef ::xsd::cxx::tree::base64_binary< " <<
            char_type << " > base64_binary;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::HexBinary&)
        {
          os << "typedef ::xsd::cxx::tree::hex_binary< " <<
            char_type << " > hex_binary;"
             << endl;
        }


        // Date/time.
        //
        virtual Void
        traverse (SemanticGraph::Fundamental::Date&)
        {
          os << "// Date/time." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::tree::date< " <<
            char_type << " > date;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::DateTime&)
        {
          os << "typedef ::xsd::cxx::tree::date_time< " <<
            char_type << " > date_time;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Duration&)
        {
          os << "typedef ::xsd::cxx::tree::duration< " <<
            char_type << " > duration;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Day&)
        {
          os << "typedef ::xsd::cxx::tree::day< " <<
            char_type << " > day;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Month&)
        {
          os << "typedef ::xsd::cxx::tree::month< " <<
            char_type << " > month;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::MonthDay&)
        {
          os << "typedef ::xsd::cxx::tree::month_day< " <<
            char_type << " > month_day;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Year&)
        {
          os << "typedef ::xsd::cxx::tree::year< " <<
            char_type << " > year;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::YearMonth&)
        {
          os << "typedef ::xsd::cxx::tree::year_month< " <<
            char_type << " > year_month;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Time&)
        {
          os << "typedef ::xsd::cxx::tree::time< " <<
            char_type << " > time;"
             << endl;
        }

        // Entity.
        //
        virtual Void
        traverse (SemanticGraph::Fundamental::Entity&)
        {
          os << "// Entity." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::tree::entity< " <<
            char_type << " > entity;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Entities&)
        {
          os << "typedef ::xsd::cxx::tree::entities< " <<
            char_type << " > entities;"
             << endl;
        }

        virtual Void
        post (SemanticGraph::Namespace& n)
        {
          os << "// Exceptions." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::tree::exception< " <<
            char_type << " > exception;"
             << "typedef ::xsd::cxx::tree::parsing< " <<
            char_type << " > parsing;"
             << "typedef ::xsd::cxx::tree::expected_element< " <<
            char_type << " > expected_element;"
             << "typedef ::xsd::cxx::tree::unexpected_element< " <<
            char_type << " > unexpected_element;"
             << "typedef ::xsd::cxx::tree::expected_attribute< " <<
            char_type << " > expected_attribute;"
             << "typedef ::xsd::cxx::tree::unexpected_enumerator< " <<
            char_type << " > unexpected_enumerator;"
             << "typedef ::xsd::cxx::tree::no_type_info< " <<
            char_type << " > no_type_info;"
             << "typedef ::xsd::cxx::tree::not_derived< " <<
            char_type << " > not_derived;"
             << "typedef ::xsd::cxx::tree::duplicate_id< " <<
            char_type << " > duplicate_id;"
             << "typedef ::xsd::cxx::tree::serialization< " <<
            char_type << " > serialization;"
             << "typedef ::xsd::cxx::tree::no_namespace_mapping< " <<
            char_type << " > no_namespace_mapping;"
             << "typedef ::xsd::cxx::tree::no_prefix_mapping< " <<
            char_type << " > no_prefix_mapping;"
             << "typedef ::xsd::cxx::tree::xsi_already_in_use< " <<
            char_type << " > xsi_already_in_use;"
             << "typedef ::xsd::cxx::tree::bounds< " <<
            char_type << " > bounds;"
             << endl;

          os << "// Parsing/serialization error." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::tree::error< " <<
            char_type << " > error;"
             << "typedef ::xsd::cxx::tree::errors< " <<
            char_type << " > errors;"
             << endl;

          os << "// Error handler interface." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::xml::error_handler< " <<
            char_type << " > error_handler;"
             << endl;

          if (options.value<CLI::generate_serialization> ())
          {
            os << "// Namespace information. Used in serialization functions." << endl
               << "//" << endl
               << "typedef ::xsd::cxx::xml::dom::namespace_info < " <<
              char_type << " > namespace_info;"
               << "typedef ::xsd::cxx::xml::dom::namespace_infomap < " <<
              char_type << " > namespace_infomap;"
               << endl;
          }

          os << "// Flags and properties." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::tree::flags flags;"
             << "typedef ::xsd::cxx::tree::properties< " << char_type <<
            " > properties;"
             << endl;

          os << "// DOM user data key for back pointers to tree nodes." << endl
             << "//" << endl;

          String g (L"XSD_CXX_TREE_TREE_NODE_KEY_IN_" + ns_name (n));

          std::transform (g.begin (), g.end(), g.begin (), upcase);
          g = escape (g); // Make it a C++ id.

          os << "#ifndef " << g << endl
             << "#define " << g << endl
             << endl
             << "XMLCh const* const tree_node_key = " <<
            "::xsd::cxx::tree::user_data_keys::node;"
             << endl
             << "#endif" << endl;

          Namespace::post (n);
        }

      private:
        Traversal::Names names_;

      };

      // IDREF/IDREFS mapping.
      //
      struct IdRef : Traversal::Fundamental::IdRef,
                     Traversal::Fundamental::IdRefs,
                     protected virtual Context
      {
        IdRef (Context& c, String const& name_ = L"")
            : Context (c),
              name (name_)
        {
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::IdRef& i)
        {
          // IdRef specialization is never named.
          //
          os << "typedef ::xsd::cxx::tree::idref< " <<
            type_name (i.argumented ().type ()) << ", " <<
            char_type << " > " << name << ";";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::IdRefs& i)
        {
          // IdRefs specialization is never named.
          //
          os << "typedef ::xsd::cxx::tree::idrefs< " <<
            type_name (i.argumented ().type ()) << ", " <<
            char_type << " > " << name << ";";
        }

      private:
        String
        type_name (SemanticGraph::Type& t)
        {
          // This type is always named.
          //
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (t);

          return o.str ();
        }

      private:
        String name;
      };


      // List mapping.
      //
      struct List : Traversal::List, protected virtual Context
      {
        List (Context& c, String const& name_ = L"")
            : Context (c),
              name (name_),
              ex (c.esymbol.empty () ? c.esymbol : c.esymbol + L" ")
        {
        }

        virtual Void
        traverse (Type& l)
        {
          if (l.named ())
            name = ename (l);

          os << "class " << ex << name <<
            ": public ::xsd::cxx::tree::simple_type, " <<
            "public ::xsd::cxx::tree::sequence< " <<
            type_name (l.argumented ().type ()) << " >"
             << "{"
             << "public:" << endl;

          // c-tor ()
          //
          os << name << " ();"
             << endl;

          // c-tor (xml::element const&)
          //
          os << name << " (" << xml_element_type << " const&," << endl
             << "::xsd::cxx::tree::flags," << endl
             << "::xsd::cxx::tree::type*);"
             << endl;

          // c-tor (xml::attribute const&)
          //
          os << name << " (" << xml_attribute_type << " const&," << endl
             << "::xsd::cxx::tree::flags," << endl
             << "::xsd::cxx::tree::type*);"
             << endl;

          // c-tor (std::basic_string const&, xml::element const*)
          //
          os << name << " (" << string_type << " const&," << endl
             << xml_element_type << " const*," << endl
             << "::xsd::cxx::tree::flags," << endl
             << "::xsd::cxx::tree::type*);"
             << endl;

          // copy c-tor ()
          //
          os << name << " (" << name << " const&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << "::xsd::cxx::tree::type* = 0);"
             << endl;

          os << "virtual " << name << "*" << endl
             << "_clone (::xsd::cxx::tree::flags = 0," << endl
             << "::xsd::cxx::tree::type* = 0) const;"
             << endl;

          os << "};";
        }

      private:
        String
        type_name (SemanticGraph::Type& t)
        {
          // This type is always named.
          //
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (t);

          return o.str ();
        }

      private:
        String name;
        String ex;
      };


      // Union mapping.
      //
      struct Union : Traversal::Union, protected virtual Context
      {
        Union (Context& c, String const& name_ = L"")
            : Context (c),
              name (name_),
              ex (c.esymbol.empty () ? c.esymbol : c.esymbol + L" ")
        {
        }

        virtual Void
        traverse (Type& u)
        {
          if (u.named ())
            name = ename (u);

          os << "class " << ex << name <<
            ": public ::xsd::cxx::tree::string< " << char_type << " >"
             << "{"
             << "public:" << endl
             << endl;

          // c-tor (string const&)
          //
          os << name << " (" << string_type << " const&);"
             << endl;

          // c-tor (xml::element const&)
          //
          os << name << " (" << xml_element_type << " const&," << endl
             << "::xsd::cxx::tree::flags," << endl
             << "::xsd::cxx::tree::type*);"
             << endl;

          // c-tor (xml::attribute const&)
          //
          os << name << " (" << xml_attribute_type << " const&," << endl
             << "::xsd::cxx::tree::flags," << endl
             << "::xsd::cxx::tree::type*);"
             << endl;

          // c-tor (std::basic_string const&, xml::element const*)
          //
          os << name << " (" << string_type << " const&," << endl
             << xml_element_type << " const*," << endl
             << "::xsd::cxx::tree::flags," << endl
             << "::xsd::cxx::tree::type*);"
             << endl;

          // copy c-tor ()
          //
          os << name << " (" << name << " const&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << "::xsd::cxx::tree::type* = 0);"
             << endl;

          os << "virtual " << name << "*" << endl
             << "_clone (::xsd::cxx::tree::flags = 0," << endl
             << "::xsd::cxx::tree::type* = 0) const;"
             << endl;

          os << "};";
        }

      private:
        String name;
        String ex;
      };

      // Enum mapping.
      //
      struct Enumerator : Traversal::Enumerator, protected virtual Context
      {
        Enumerator (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          os << ename (e);
        }
      };


      struct Enumeration : Traversal::Enumeration, protected virtual Context
      {
        Enumeration (Context& c,
                     Traversal::NodeDispatcherBase& complex,
                     String const& name_ = L"")
            : Context (c),
              complex_ (complex),
              name (name_),
              ex (c.esymbol.empty () ? c.esymbol : c.esymbol + L" "),
              base_ (c),
              member_ (c),
              enumerator_ (c)
        {
          inherits_base_ >> base_;
          inherits_member_ >> member_;

          names_ >> enumerator_;
        }

        virtual Void
        traverse (Type& e)
        {
          if (e.named ())
            name = ename (e);

          Boolean string_based (false);
          {
            IsStringBasedType t (string_based);
            t.dispatch (e);
          }

          UnsignedLong enum_count (0);

          for (Type::NamesIterator i (e.names_begin ()), end (e.names_end ());
               i != end; ++i)
            ++enum_count;

          os << "class " << ex << name << ": public ";

          // Enumeration always has a base.
          //
          inherits (e, inherits_base_);

          os << "{"
             << "public:" << endl;

          if (string_based)
          {
            os << "enum _xsd_" << name
               << "{";

            enter_scope (name);

            names<Enumeration> (e, names_, 0, 0, 0, &Enumeration::comma);

            leave_scope ();

            os << "};";
          }

          // default c-tor
          //
          if (options.value<CLI::generate_default_ctor> ())
          {
            os << name << " ();"
               << endl;
          }

          // c-tor (_xsd_name)
          //
          if (string_based)
          {
            os << name << " (_xsd_" << name << ");"
               << endl;
          }

          // c-tor (base)
          //
          os << name << " (";

          inherits (e, inherits_member_);

          os << " const&);"
             << endl;

          // c-tor (xml::element const&)
          //
          os << name << " (" << xml_element_type << " const&," << endl
             << "::xsd::cxx::tree::flags," << endl
             << "::xsd::cxx::tree::type*);"
             << endl;

          // c-tor (xml::attribute const&)
          //
          os << name << " (" << xml_attribute_type << " const&," << endl
             << "::xsd::cxx::tree::flags," << endl
             << "::xsd::cxx::tree::type*);"
             << endl;

          // c-tor (std::basic_string const&, xml::element const&)
          //
          os << name << " (" << string_type << " const&," << endl
             << xml_element_type << " const*," << endl
             << "::xsd::cxx::tree::flags," << endl
             << "::xsd::cxx::tree::type*);"
             << endl;

          // copy c-tor
          //
          os << name << " (" << name << " const&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << "::xsd::cxx::tree::type* = 0);"
             << endl;

          os << "virtual " << name << "*" << endl
             << "_clone (::xsd::cxx::tree::flags = 0," << endl
             << "::xsd::cxx::tree::type* = 0) const;"
             << endl;

          // operator= (_xsd_name)
          //
          if (string_based)
          {
            os << name << "&" << endl
               << "operator= (_xsd_" << name << ");"
               << endl;
          }

          // operator _xsd_name ()
          //
          // Lookup differences in various compilers and some really
          // twisted anonymous type hierarchies make generation of this
          // operator outside of the class a really hard task. So we are
          // going to make it "always inline".
          //
          if (string_based)
          {
            os << "operator _xsd_" << name << " () const"
               << "{"
               << "return _xsd_" << name << "_conver ();"
               << "}";
          }

          //
          //
          if (string_based)
          {
            os << "private:" << endl
               << "_xsd_" << name << endl
               << "_xsd_" << name << "_conver () const;"
               << endl
               << "static " << char_type << " const* const " <<
              "_xsd_" << name << "_literals_[" << enum_count << "];"
               << "static _xsd_" << name << " const " <<
              "_xsd_" << name << "_indexes_[" << enum_count << "];";
          }

          os << "};";
        }

        virtual Void
        comma (Type&)
        {
          os << "," << endl;
        }

      private:
        Traversal::NodeDispatcherBase& complex_;

        String name;
        String ex;

        Traversal::Inherits inherits_base_;
        BaseTypeName base_;

        Traversal::Inherits inherits_member_;
        MemberTypeName member_;

        Traversal::Names names_;
        Enumerator enumerator_;
      };


      //
      //
      struct MemberForward : Traversal::Member, protected virtual Context
      {
        MemberForward (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));

          if (name == L"type")
          {
            os << "// Workaround for g++ bug# 23206." << endl
               << "//" << endl
               << "public:" << endl
               << "struct " << name << ";"
               << endl;
          }
        }
      };


      // List of using declaration for all non-optional members.
      //
      struct UsingDecl : Traversal::Complex,
                         Traversal::Member,
                         protected virtual Context
      {
        UsingDecl (Context& c, String const& name)
            : Context (c), first_ (true), name_ (name)
        {
          inherits_ >> *this;
          names_ >> *this;
        }

        ~UsingDecl ()
        {
          if (!first_)
            os << "#endif" << endl
               << endl;
        }

        virtual Void
        traverse (SemanticGraph::Complex& c)
        {
          inherits (c, inherits_);

          enter_scope (ename (c));
          names (c, names_);
          leave_scope ();
        }

        virtual Void
        traverse (SemanticGraph::Member& m)
        {
          if (skip (m)) return;

          // Note that we are not going to include attributes with
          // default values here because they are going to be default-
          // initialized.
          //
          if (min (m) == 1 && max (m) == 1)
          {
            if (first_)
            {
              os << "#if defined(__EDG_VERSION__) || " <<
                "(defined(__HP_aCC) && __HP_aCC >= 60000)" << endl
                 << "public:" << endl;

              first_ = false;
            }

            os << "using " << name_ << "::base_::" << ename (m) << ";";
          }
        }

      private:
        Boolean first_;
        String name_;

      private:
        Traversal::Inherits inherits_;
        Traversal::Names names_;
      };


      //
      //
      struct MemberFunction: Traversal::Member, protected virtual Context
      {
        MemberFunction (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (skip (m)) return;

          String const& name (ename (m));

          if (max (m) != 1)
          {
            // sequence
            //
            String container (iscope_name (name, "container"));

            // container const&
            // name () const;
            //
            os << container << " const&" << endl
               << name << " () const;"
               << endl;

            // container&
            // name ();
            //
            os << container << "&" << endl
               << name << " ();"
               << endl;

            // void
            // name (container const&);
            //
            os << "void" << endl
               << name << " (" << container << " const&);"
               << endl;
          }
          else if (min (m) == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
            String type (iscope_name (name, "type"));
            String container (iscope_name (name, "container"));

            // container const&
            // name () const;
            //
            os << container << " const&" << endl
               << name << " () const;"
               << endl;

            // container&
            // name ();
            //
            os << container << "&" << endl
               << name << " ();"
               << endl;

            // void
            // name (type const&);
            //
            os << "void" << endl
               << name << " (" << type << " const&);"
               << endl;

            // void
            // name (container const&);
            //
            os << "void" << endl
               << name << " (" << container << " const&);"
               << endl;

            // void
            // name (auto_ptr<type>);
            //
            os << "void" << endl
               << name << " (::std::auto_ptr< " << type << " >);"
               << endl;
          }
          else
          {
            // one
            //
            String type (iscope_name (name, "type"));

            // type const&
            // name () const;
            //
            os << type << " const&" << endl
               << name << " () const;"
               << endl;

            // type&
            // name ();
            //
            os << type << "&" << endl
               << name << " ();"
               << endl;

            // void
            // name (type const&);
            //
            os << "void" << endl
               << name << " (" << type << " const&);"
               << endl;

            // void
            // name (auto_ptr<type>);
            //
            os << "void" << endl
               << name << " (::std::auto_ptr< " << type << " >);"
               << endl;
          }
        }
      };

      template <typename Complex>
      struct Member : Traversal::Member, protected virtual Context
      {
        Member (Context& c)
            : Context (c),
              type_name_ (c),
              member_function_ (c)
        {
          belongs_ >> type_name_;
        }

        virtual Void
        traverse (Type& m)
        {
          if (skip (m)) return;

          String const& name (ename (m));

          os << "// " << m.name () << endl
             << "// " << endl;


          // For mapping.
          //
          os << "public:" << endl
             << "struct " << name
             << "{";

          SemanticGraph::Type& t (m.type ());

          if (t.named ())
          {
            os << "typedef ";

            belongs (m, belongs_);

            os << " " << unclash (name, "type") << ";";
          }
          else
          {
            if (t.context ().count ("seen"))
            {
              os << "typedef " <<
                t.context ().template get<String> ("anon-name") <<
                " " << unclash (name, "type") << ";";
            }
            else
            {
              os << "struct _xsd_" << name << "_"
                 << "{";

              Traversal::Belongs belongs_;

              IdRef id_ref (*this, name);
              List list (*this, name);
              Union union_ (*this, name);
              Complex complex (*this, name);
              Enumeration enumeration (*this, complex, name);

              belongs_ >> id_ref;
              belongs_ >> list;
              belongs_ >> union_;
              belongs_ >> complex;
              belongs_ >> enumeration;

              enter_scope (name);
              enter_scope (L"_xsd_" + name + L"_");

              t.context ().set ("seen", true);

              belongs (m, belongs_);

              t.context ().remove ("seen");

              leave_scope ();
              leave_scope ();

              os << "};";

              os << "typedef _xsd_" << name << "_::" <<
                name << " " << unclash (name, "type") << ";";
            }
          }

          os << "typedef ::xsd::cxx::tree::traits< " <<
            unclash (name, "type") << " > " << unclash (name, "traits") <<
            ";";

          if (max (m) != 1)
          {
            // sequence
            //
            os << "typedef ::xsd::cxx::tree::sequence< " <<
              unclash (name, "type") << " > " <<
              unclash (name, "container") << ";"

               << "typedef " << unclash (name, "container") <<
              "::iterator " << unclash (name, "iterator") << ";"

               << "typedef " << unclash (name, "container") <<
              "::const_iterator " << unclash (name, "const_iterator") << ";";

          }
          else if (min (m) == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
            os << "typedef ::xsd::cxx::tree::optional< " <<
              unclash (name, "type") << " > " <<
              unclash (name, "container") << ";";
          }
          else
          {
            // one
            //
          }


          // default_value
          //
          if (m.default_ ())
          {
            Boolean simple (true);

            if (m.is_a<SemanticGraph::Element> ())
            {
              IsSimpleType test (simple);
              test.dispatch (t);
            }

            if (simple)
            {
              os << endl
                 << "static " << unclash (name, "type") << " const&" << endl
                 << unclash (name, "default_value") << " ();"
                 << endl
                 << "private:" << endl
                 << "static " << unclash (name, "type") << " const " <<
                unclash (name, "default_value") << "_;";
            }
          }

          os << "};";

          member_function_.traverse (m);
        }

      private:
        MemberTypeName type_name_;
        Traversal::Belongs belongs_;

        MemberFunction member_function_;
      };



      struct DataMember: Traversal::Member, protected virtual Context
      {
        DataMember (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (skip (m)) return;

          String const& name (ename (m));

          if (max (m) != 1)
          {
            // sequence
            //
            os << "::xsd::cxx::tree::sequence_functor< " <<
              iscope_name (name, "type") << " > " <<
              "_xsd_" << name << "_;";
          }
          else if (min (m) == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
            os << "::xsd::cxx::tree::optional_functor< " <<
              iscope_name (name, "type") << " > " <<
              "_xsd_" << name << "_;";
          }
          else
          {
            // one
            //
            os << "::xsd::cxx::tree::one_functor< " <<
              iscope_name (name, "type") << " > " <<
              "_xsd_" << name << "_;";
          }
        }
      };


      struct Complex : Traversal::Complex, protected virtual Context
      {
        Complex (Context& c, String const& name_ = L"")
            : Context (c),
              name (name_),
              e (c.esymbol.empty () ? c.esymbol : c.esymbol + L" "),
              base_name_ (c),
              member_name_ (c),
              member_forward_ (c),
              member_ (c),
              data_member_ (c),
              from_base_ctor_arg_ (c, false)
        {
          inherits_base_ >> base_name_;
          inherits_member_ >> member_name_;

          names_ >> member_;
          names_data_ >> data_member_;
          names_forward_ >> member_forward_;

          from_base_ctor_arg_names_ >> from_base_ctor_arg_;
        }

        virtual Void
        traverse (Type& c)
        {
          if (c.named ())
            name = ename (c);

          Boolean has_members (has<Traversal::Member> (c));

          Boolean simple (true);
          {
            IsSimpleType t (simple);
            t.dispatch (c);
          }

          os << "class " << e << name << ": public ";

          inherits<Complex> (c,
                             inherits_base_,
                             0, 0, &Complex::inherits_none);

          os << "{"
             << "public:" << endl
             << endl
             << "struct _xsd_" << name
             << "{"
             << "typedef ";

          inherits<Complex> (c,
                             inherits_member_,
                             0, 0, &Complex::inherits_none);

          os << " base_;"

             << "};";

          enter_scope (name);
          c.context ().set ("anon-name", scope);

          // Forward declarations.
          //
          names (c, names_forward_);


          //
          //
          names (c, names_);

          c.context ().remove ("anon-name");
          leave_scope ();


          // Using declarations for all non-optional base members. They are
          // needed for EDG frontent-based compilers (Intel C++, aCC6, etc).
          //
          {
            UsingDecl complex (*this, L"_xsd_" + name);
            Traversal::Inherits inherits_complex;

            inherits_complex >> complex;

            inherits (c, inherits_complex);
          }

          os << "// Constructors." << endl
             << "//" << endl
             << "public:" << endl;

          // default c-tor
          //
          if (options.value<CLI::generate_default_ctor> ())
          {
            // c-tor (all-non-optional-members) will become default
            // c-tor if our inheritance hierarchy has no non-optional
            // members and no simple base.
            //
            Boolean generate (false);
            {
              GenerateDefaultCtor t (*this, generate);
              t.traverse (c);
            }

            if (generate)
            {
              os << name << " ();"
                 << endl;
            }
          }

          // c-tor (base, all-non-optional-members)
          //
          if (options.value<CLI::generate_from_base_ctor> ())
          {
            // c-tor (all-non-optional-members) will become
            // c-tor (base, all-non-optional-members) unless our
            // immediate base's hierarchy has some non-optional
            // members.
            //
            Boolean generate (false);
            {
              GenerateFromBaseCtor t (*this, generate);
              t.traverse (c);
            }

            if (generate)
            {
              String base_member (L"_xsd_" + name + L"::base_");

              os << name << " (" << base_member << " const&";

              enter_scope (name);
              names (c, from_base_ctor_arg_names_);
              leave_scope ();

              os << ");"
                 << endl;
            }
          }

          // c-tor (all-non-optional-members)
          //
          os << name << " (";

          {
            enter_scope (name);

            CtorArgs ctor_args (*this, L"_xsd_" + name, false);
            ctor_args.dispatch (c);

            leave_scope ();
          }

          os << ");"
             << endl;

          // c-tor (xml::element const&)
          //
          os << name << " (" << xml_element_type << " const&," << endl
             << "::xsd::cxx::tree::flags," << endl
             << "::xsd::cxx::tree::type*);"
             << endl;


          if (simple)
          {
            // c-tor (xml::attribute const&)
            //
            os << name << " (" << xml_attribute_type << " const&," << endl
               << "::xsd::cxx::tree::flags," << endl
               << "::xsd::cxx::tree::type*);"
               << endl;

            // c-tor (std::basic_string const&, xml::element const*)
            //
            os << name << " (" << string_type << " const&," << endl
               << xml_element_type << " const*," << endl
               << "::xsd::cxx::tree::flags," << endl
               << "::xsd::cxx::tree::type*);"
               << endl;
          }

          // copy c-tor
          //
          os << name << " (" << name << " const&," << endl
             << "::xsd::cxx::tree::flags = 0," << endl
             << "::xsd::cxx::tree::type* = 0);"
             << endl;

          os << "virtual " << name << "*" << endl
             << "_clone (::xsd::cxx::tree::flags = 0," << endl
             << "::xsd::cxx::tree::type* = 0) const;"
             << endl;


          // Data members.
          //
          if (has_members)
          {
            os << "// Implementation." << endl
               << "//" << endl
               << "private:" << endl;


            // parse(xml::element const&)
            //
            os << "void" << endl
               << unclash (name, "parse") << " (" <<
              xml_element_type << " const&," << endl
               << "::xsd::cxx::tree::flags);"
               << endl;


            enter_scope (name);
            names (c, names_data_);
            leave_scope ();
          }

          os << "};";

        }

        virtual Void
        inherits_none (Type& t)
        {
          os << "::xsd::cxx::tree::type";
        }

      private:
        String name;
        String e;

        Traversal::Inherits inherits_base_;
        Traversal::Inherits inherits_member_;

        BaseTypeName base_name_;
        MemberTypeName member_name_;

        Traversal::Names names_forward_;
        MemberForward member_forward_;

        Traversal::Names names_;
        Member<Complex> member_;

        Traversal::Names names_data_;
        DataMember data_member_;

        FromBaseCtorArg from_base_ctor_arg_;
        Traversal::Names from_base_ctor_arg_names_;
      };


      // Goes after anonymous types via global element. Note that we don't
      // need to go after attributes.
      //
      struct GlobalMember: Traversal::Element,
                           GlobalElementBase,
                           protected virtual Context
      {
        GlobalMember (Context& c)
            : GlobalElementBase (c), Context (c), type_name_ (c)
        {
          belongs_ >> type_name_;
        }

        virtual Void
        traverse (Type& m)
        {
          if (!generate_p (m))
            return;

          String const& name (ename (m));
          SemanticGraph::Type& t (m.type ());

          if (!t.named ())
          {
            os << "// " << m.name () << endl
               << "// " << endl;

            os << "class " << name
               << "{"
               << "struct _xsd_" << name << "_"
               << "{";

            Traversal::Belongs belongs_;

            IdRef id_ref (*this, name);
            List list (*this, name);
            Union union_ (*this, name);
            Complex complex (*this, name);
            Enumeration enumeration (*this, complex, name);

            belongs_ >> id_ref;
            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();

            os << "};"
               << "public:" << endl
               << "typedef _xsd_" << name << "_::" << name << " "
               << unclash (name, "type") << ";"
               << "};";
          }
        }

      private:
        MemberTypeName type_name_;
        Traversal::Belongs belongs_;
      };


      struct Includes : Traversal::Imports,
                        Traversal::Includes,
                        protected virtual Context
      {
        Includes (Context& c, String const& expr)
            : Context (c), expr_ (expr)
        {
        }

        virtual Void
        traverse (SemanticGraph::Imports& i)
        {
          traverse (i.path ());
        }

        virtual Void
        traverse (SemanticGraph::Includes& i)
        {
          traverse (i.path ());
        }

        virtual Void
        traverse (SemanticGraph::Path const& path)
        {
          using BackendElements::Regex::perl_s;

          // Note the use of the portable representation of the path.
          //
          String name (perl_s (String (path.string ()), expr_));

          Char op (include_with_brackets ? '<' : '"');
          Char cl (include_with_brackets ? '>' : '"');

          os << "#include " << op << include_prefix << name << cl << endl
             << endl;
        }

      private:
        String expr_;
      };
    }

    Void
    generate_tree_header (Context& ctx,
                          SemanticGraph::Schema& tu,
                          String const& hxx_expr,
                          String const& ixx_expr)
    {
      String c (ctx.char_type);

      if (c == L"char")
      {
        ctx.os << "#ifndef XSD_USE_CHAR" << endl
               << "#define XSD_USE_CHAR" << endl
               << "#endif" << endl
               << endl;

        ctx.os << "#ifndef XSD_CXX_TREE_USE_CHAR" << endl
               << "#define XSD_CXX_TREE_USE_CHAR" << endl
               << "#endif" << endl
               << endl;
      }
      else if (c == L"wchar_t")
      {
        ctx.os << "#ifndef XSD_USE_WCHAR" << endl
               << "#define XSD_USE_WCHAR" << endl
               << "#endif" << endl
               << endl;

        ctx.os << "#ifndef XSD_CXX_TREE_USE_WCHAR" << endl
               << "#define XSD_CXX_TREE_USE_WCHAR" << endl
               << "#endif" << endl
               << endl;
      }

      ctx.os << "#include <memory>    // std::auto_ptr" << endl
             << "#include <algorithm> // std::binary_search" << endl
             << endl
             << "#include <xsd/cxx/xml/error-handler.hxx>" << endl
             << "#include <xsd/cxx/xml/dom/elements.hxx>" << endl;

      if (ctx.options.value<CLI::generate_serialization> ())
        ctx.os << "#include <xsd/cxx/xml/dom/serialization.hxx>" << endl;


      ctx.os << endl
             << "#include <xsd/cxx/tree/exceptions.hxx>" << endl
             << "#include <xsd/cxx/tree/elements.hxx>" << endl
             << "#include <xsd/cxx/tree/containers.hxx>" << endl
             << "#include <xsd/cxx/tree/functors.hxx>" << endl
             << "#include <xsd/cxx/tree/types.hxx>" << endl
             << endl;

      // Emit fundamental types.
      //
      {
        Traversal::Schema schema, xsd;
        Traversal::Implies implies;
        Traversal::Names names;
        FundamentalNamespace ns (ctx);

        schema >> implies >> xsd >> names >> ns;

        schema.dispatch (tu);
      }

      // Emit header includes.
      //
      {
        if (ctx.options.value<CLI::generate_inline> ())
        {
          ctx.os << "#ifndef XSD_DONT_INCLUDE_INLINE" << endl
                 << "#define XSD_DONT_INCLUDE_INLINE" << endl
                 << endl;
        }

        Traversal::Schema schema;
        Includes includes (ctx, hxx_expr);

        schema >> includes;

        schema.dispatch (tu);

        if (ctx.options.value<CLI::generate_inline> ())
        {
          ctx.os << "#undef XSD_DONT_INCLUDE_INLINE" << endl
                 << "#else" << endl
                 << endl;

          Traversal::Schema schema;
          Includes includes (ctx, hxx_expr);

          schema >> includes;

          schema.dispatch (tu);

          ctx.os << "#endif // XSD_DONT_INCLUDE_INLINE" << endl
                 << endl;
        }
      }


      {
        Traversal::Schema schema;

        Traversal::Sources sources;
        Traversal::Names names_ns, names;

        Namespace ns (ctx, true);

        List list (ctx);
        Union union_ (ctx);
        Complex complex (ctx);
        Enumeration enumeration (ctx, complex);
        GlobalMember member (ctx);

        schema >> sources >> schema;
        schema >> names_ns >> ns >> names;

        // Note that there is no IdRef since it is always anonymous.
        //
        names >> list;
        names >> union_;
        names >> complex;
        names >> enumeration;
        names >> member;

        schema.dispatch (tu);
      }

      // Emit inline includes.
      //
      if (ctx.options.value<CLI::generate_inline> ())
      {
        ctx.os << "#ifndef XSD_DONT_INCLUDE_INLINE" << endl
               << endl;

        Traversal::Schema schema;
        Includes includes (ctx, ixx_expr);

        schema >> includes;

        schema.dispatch (tu);

        ctx.os << "#endif // XSD_DONT_INCLUDE_INLINE" << endl
               << endl;
      }
    }
  }
}
